/**
 *
 * @file main.cc
 * @author Lasse Lehtonen
 *
 * Purpose: Transaction Generator (TG) models a multiprocessor SoC. 
 *          System and workload are described in an XML file. 
 *          This is a command line tool but there is option to 
 *          use visualization program called ExecutioMonitor.
 *
 */

/*
 * Copyright 2010 Tampere University of Technology
 * 
 *  This file is part of Transaction Generator.
 *
 *  Transaction Generator is free software: you can redistribute it
 *  and/or modify it under the terms of the Lesser GNU General Public
 *  License as published by the Free Software Foundation, either
 *  version 3 of the License, or (at your option) any later version.
 *
 *  Transaction Generator is distributed in the hope that it will be
 *  useful, but WITHOUT ANY WARRANTY; without even the implied
 *  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the Lesser GNU General Public License for more details.
 *
 *  You should have received a copy of the Lesser GNU General Public
 *  License along with Transaction Generator.  If not, see
 *  <http://www.gnu.org/licenses/>.
 */

/*
 * $Id: main.cc 1951 2011-08-19 12:43:09Z ege $
 *
 */


#include "amount.hh"
#include "event.hh"
#include "configuration.hh"
#include "resource.hh"
#include "processing_element.hh"
#include "task.hh"
#include "mem_area.hh"
#include "trigger.hh"
#include "measure.hh"
#include "noc_factory.hh"
#include "memory_model.hh"
#include "cost_function.hh"

// TCP connection to Execution Monitor cannot be used in Modelsim. 
// This causes several Ifdef/ifndef later on as well. 
// Variable MTI_SYSTEMC (as in Model Technology Inc.) is set by compiler

#ifndef MTI_SYSTEMC
#include <boost/program_options.hpp>
#ifdef SCTG_USE_EXECMON
#include "non_sc_factory.hh"
#include "tcp_server.hh"
#include <boost/asio.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>
#endif
#endif

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <string>
#include <set>   
#include <exception>
#include <ctime> 
#include <vector>
#include <map> 

#include <boost/foreach.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/xml_parser.hpp>
#include <boost/random.hpp>
#include <boost/lexical_cast.hpp>

#include <systemc> 


void analyzeModel (sctg::Configuration*                   cfg,
		 std::vector<sctg::ProcessingElement*>* pes,
		 std::vector<sctg::Task*>*              tasks,
		 std::vector<sctg::Event*>*             events,
		 std::vector<sctg::MemoryModel*>*       mems,
		 std::vector<sctg::MemArea*>*           memAreas,
		 boost::property_tree::ptree*           tgModel

   ); 

void calculateCost (sctg::Configuration* cfg, sctg::CostFunction* costF,
		    std::vector<sctg::ProcessingElement*>* pes,
		    std::vector<sctg::Task*>*              tasks,
		    std::vector<sctg::Event*>*             events
   ); 
void printCost (sctg::Configuration* cfg, sctg::CostFunction* costF);


/** Main function, daa
 */
int sc_main(int argc, char* argv[])
{    
   std::vector<sctg::ProcessingElement*> pes;
   std::vector<sctg::MemoryModel*>       mems;
   std::vector<sctg::Task*>              tasks;
   std::vector<sctg::MemArea*>           memAreas;
   std::vector<sctg::Event*>             events;      
   NocFactory*                           nocFactory = 0;
   sctg::Measure*                        measurements = 0;
   sctg::Configuration*                  conf = 0;
   
   sctg::CostFunction                    costFunctionParser;

   using boost::property_tree::ptree;
   ptree tgModel;
   ptree peLib;

   unsigned int playbackDelay = 0;

   enum ProgramMode {NORMAL, PLAYBACK, MIXEDSIM};
   ProgramMode programMode = NORMAL;

   bool useExecutionMonitor = false;

#ifdef SCTG_USE_EXECMON
#ifndef MTI_SYSTEMC
   sctg::NonScFactory* nonScFactory;
#endif
#endif      
  
   try
   {
      std::cout << std::endl
		<< "Transaction Generator 2" << std::endl;

      std::string inputFile;
      std::string playbackFile;
      std::string saveDir;

      // There 2 main phases and few sub-phases

      // 1. Handle command line parameters 
      //    a) with Boost, b) simpler version for Modelsim

      // 2. Execute phase 
      // 2.a) Playback a log file in ExecMon
      // 2.b) Run simulation 
      // 2.b.1. Read, validate XML, and optionally connect to ExecMon 
      // 2.b.2. Run simulation
      // 2.b.3) Report statistics and delete dynamicalle reserved memory


#ifndef MTI_SYSTEMC

      // -----------
      //  Phase 1.a) Command line option parser using Boost.ProgramOptions
      // -----------
      
      namespace po = boost::program_options;
      po::options_description poDesc("Allowed options");
      poDesc.add_options()
	 ("help", "This message")
	 ("input-file,i", po::value<std::string>(), "Input XML file")
	 ("save-directory,s", po::value<std::string>(), 
	  "Directory to save logs")
	 ("execution-monitor,e", "Use Execution Monitor")
	 ("replay-file,r", po::value<std::string>(), 
	  "Recorded log file to replay in ExecMon")
	 ("delay,d", po::value<unsigned int>(), 
	  "Delay (ms) to slow down replay");
      
      po::variables_map poVm;
      po::store(po::parse_command_line(argc, argv, poDesc), poVm);
      po::notify(poVm);
      
      if(poVm.count("help"))
      {
	 std::cout << poDesc << std::endl;
	 return EXIT_SUCCESS;
      }

      if(poVm.count("input-file"))
      {
	 std::cout << "input-file:             " 
		   << poVm["input-file"].as<std::string>()
		   << std::endl;
	 inputFile = poVm["input-file"].as<std::string>();
	 programMode = NORMAL;

	 if(poVm.count("execution-monitor"))
	 {
#ifndef SCTG_USE_EXECMON
	    std::cerr << "You need to compile Transaction Generator 2 "
		      << "with SCTG_USE_EXECMON flag to use Execution Monitor"
		      << std::endl;
	    return EXIT_FAILURE;
#endif	    
	    useExecutionMonitor = true;
	 }
      }
      else if(poVm.count("replay-file"))
      {
#ifndef SCTG_USE_EXECMON
	 std::cerr << "You need to compile Transaction Generator 2 "
		   << "with SCTG_USE_EXECMON flag to use Execution Monitor"
		   << std::endl;
	 return EXIT_FAILURE;
#endif
	 std::cout << "replay-file:           "
		   << poVm["replay-file"].as<std::string>()
		   << std::endl;
	 playbackFile = poVm["replay-file"].as<std::string>();
	 programMode = PLAYBACK;
      }
      else
      {
	 std::cout << poDesc << std::endl;
	 return EXIT_FAILURE;
      }
      
      if(poVm.count("delay"))
      {
	 std::cout << "Delay for replay:       "
		   << poVm["delay"].as<unsigned int>()
		   << " ms"
		   << std::endl;
	 playbackDelay = poVm["delay"].as<unsigned int>();
      }
      else
      {
	 std::cout << "Delay for playback:     1ms" << std::endl;
	 playbackDelay = 1;
      }

      if(poVm.count("save-directory"))
      {
	 std::cout << "Directory to save logs: "
		   << poVm["save-directory"].as<std::string>()
		   << std::endl;
	 saveDir = poVm["save-directory"].as<std::string>();
      }
      else
      {
	 std::cout << "Directory to save logs: ." << std::endl;
      }

      std::cout << std::endl;

#else 
      // -----------
      // Phase 1b). Simplified command line parser for Simulation with Modelsim
      // -----------

      if(argc == 2)
      {
	 inputFile.assign(argv[1]);
      }
      else 
      {
	 std::cout << std::endl
		   << "Input file not set" << std::endl
		   << std::endl
		   << "Usage:  vsim -sc_arg <input-file> ..." << std::endl 
		   << std::endl;
	 return EXIT_FAILURE;
      }

      programMode = MIXEDSIM;

#endif      

// ----------
// Phase 2a) Visualize previous simulation with ExecutionMonitor 
// ES: Voisko t olla erillinen funktio??
// ----------

#ifdef SCTG_USE_EXECMON
#ifndef MTI_SYSTEMC

      if(programMode == PLAYBACK)
      {
	 std::cout << "Replay mode" << std::endl;
	 sctg::TcpServer server(9990);

	 boost::asio::io_service io;
	 boost::asio::deadline_timer 
	    delay(io, boost::posix_time::milliseconds(playbackDelay));

	 std::ifstream inFile(playbackFile.c_str());
	 std::string line;

	 std::cout << "Waiting for Execution Monitor to connect" << std::endl;
	 server.waitConnection();
	 std::cout << "Execution Monitor connected" << std::endl;

	 while(std::getline(inFile, line))
	 {
	    server.send(line);
	    delay.wait();
	    delay.expires_at(delay.expires_at() + 
			     boost::posix_time::milliseconds(playbackDelay));
	 }
	 
	 std::cout << "Done" << std::endl;
	 return EXIT_SUCCESS;
      }

#endif
#endif

      //-------------
      // Phase 2.b.1) Read in the XML file 
      //   It definines application and platform model, mapping
      //   constraints etc and store it to a property_tree      
      //-------------



      std::string tgFile(inputFile.c_str());      
      read_xml(tgFile, tgModel);


      // Parse XML and store necessary data into Configuration object
      ptree node = tgModel.get_child("system");
      conf = new sctg::Configuration(node, saveDir);
      conf->useExecMon(useExecutionMonitor);

      // Analyze the model and instantiate necessary pes, memories and events. 
      // All info goes to Configuration to allow easy access later.
      analyzeModel (conf,
		  &pes, 
		  &tasks, 
		  &events, 
		  &mems, 
		  &memAreas, 
		  &tgModel); //ES

      
      
      // Create NoC (which starts 1 or more threads)
      nocFactory = new NocFactory(conf);


      // Create measurement thread
      measurements = new sctg::Measure("measurer", *conf, pes, tasks, mems, 
				       memAreas);



#ifdef SCTG_USE_EXECMON
      
      if(conf->useExecMon())
      {

#ifndef MTI_SYSTEMC
	 // Create NonScFactory
	 nonScFactory = new sctg::NonScFactory(*conf);

	 std::cout << "Waiting for Execution Monitor to connect" << std::endl;
	 conf->getTcpServer()->waitConnection();
	 std::cout << "Execution Monitor connected" << std::endl;
#endif
	 // Configure Execution Monitor
      
	 std::ostringstream oss;      
	 std::string str;
	 oss.str("");
	 oss << "<configuration>";
	 for(std::map<unsigned long int, 
		sctg::Resource*>::iterator cpuIter = 
		conf->getResourceBegin(); cpuIter != conf->getResourceEnd(); 
	     ++cpuIter)
	 {
	    std::string cpuName = (*cpuIter).second->getName();
	    unsigned long int cpuId = (*cpuIter).first;
	    oss << "<cpu name=\"" << cpuName
		<< "\" id=\"" << cpuId << "\">";
	    for(std::map<unsigned long int, 
			 unsigned long int>::iterator groupIter = 
		   conf->getGroupPeBegin(); groupIter != conf->getGroupPeEnd();
		++groupIter)
	    {
	       unsigned long int group = (*groupIter).first;
	       unsigned long int pe    = (*groupIter).second;
	       if(cpuId == pe)
	       {
		  oss << "<thread name=\"Thread_" << group << "\" id=\"" 
		      << group << "\" " << "priority=\"0\">";
		  for(std::map<unsigned long int, unsigned long int>::iterator
			 taskIter = conf->getTaskGroupBegin(); 
		      taskIter != conf->getTaskGroupEnd(); ++taskIter)
		  {
		     unsigned long int task = (*taskIter).first;
		     unsigned long int gr   = (*taskIter).second;
		     if(group == gr)
		     {
			oss << "<process name=\"" 
			    << conf->getResourceUser(task)->getName()
			    << "\" id=\"" << task << "\"/>";
		     }
		  }

		  oss << "</thread>";
	       }
	       	       
	    }
	    
	    bool ismem = false;
	    for(unsigned long int memi = 0; memi < memAreas.size(); ++memi)
	    {
		  
	       if(conf->getResourceByResourceUser(memAreas.at(memi)->getId())
		  ->getId() == cpuId)
	       {
		  if(!ismem)
		  {
		     ismem = true;
		     oss << "<thread name=\"Memory\" "
			 << "id=\"999\" priority=\"0\">";
		  }
		  oss << "<process name=\"" << memAreas.at(memi)->getName()
		      << "\" id=\"" << memAreas.at(memi)->getId() 
		      << "\"/>";
	       }
	    }
	    if(ismem)
	    {
	       oss << "</thread>";
	    }

	    oss << "</cpu>";
	 }      

	 oss << "</configuration>" << std::endl;
     
	 str = oss.str();
#ifndef MTI_SYSTEMC
	 conf->getTcpServer()->send(str);
#endif
	 if(conf->getExecMonStream())
	 {
	    **conf->getExecMonStream()
	       << oss.str();
	 }

      }

#endif // USE_EXECMON     
      
      // ------------
      // Phase 2.b.2) RUN THE SIMULATION
      // ------------
      std::cout << "Starting simulation with seed " 
		<< conf->getSeed() << std::endl;

#ifdef MTI_SYSTEMC
      sc_core::sc_start();
#else
      sc_core::sc_start(conf->getSimLength());
      sc_core::sc_start(sc_core::SC_ZERO_TIME);
#endif


      std::cout << "Simulation ends at " 
		<< sc_core::sc_time_stamp() << std::endl << std::endl;
      

      // ------------
      // Phase 2.b.3) Report statistics
      // ------------


      calculateCost (conf, &costFunctionParser, &pes, &tasks, &events); //ES
      printCost (conf, &costFunctionParser); //ES
   
   }
   catch(std::exception& e)
   {
      std::cerr << "EXCEPTION: "
		<< e.what() << std::endl;
   }
   catch(...)
   {
      std::cerr << "EXCEPTION: unknown source (not std::exception)."
		<< std::endl;
   }
   



  
   // CLEAN UP: 
  
   // Destroy all processing elements, tasks, events, NoC, and measurements
   for(unsigned int i = 0; i < pes.size(); ++i)
   {
      delete pes.at(i);
      pes.at(i) = 0;
   }
   for(unsigned int i = 0; i < tasks.size(); ++i)
   {
      delete tasks.at(i);
      tasks.at(i) = 0;
   }
   for(unsigned int i = 0; i < events.size(); ++i)
   {
      delete events.at(i);
      events.at(i) = 0;
   }  
   
   delete nocFactory; nocFactory = 0;
   delete measurements; measurements = 0;


#ifdef SCTG_USE_EXECMON
#ifndef MTI_SYSTEMC
   // Destroy NonScFactory
   delete nonScFactory; nonScFactory = 0;
#endif
#endif
   
   // Destroy configuration
   delete conf; conf = 0;

   return EXIT_SUCCESS;
} // main(...)







/**
 * Analyzes the TgModel and stores necessary information inro Cxonfiguration object. 
 * Moreover, there is separate lists for pes, tasks, events, memories and memory areas.
 * New objects will be created into them. 
 * 2011-08-19 This a new function but all the code is just moved from main to here.
 */ 

void analyzeModel (sctg::Configuration* cfg,
		 std::vector<sctg::ProcessingElement*>* pes,
		 std::vector<sctg::Task*>*              tasks,
		 std::vector<sctg::Event*>*             events,
		 std::vector<sctg::MemoryModel*>*       mems,
		 std::vector<sctg::MemArea*>*           memAreas,
		 boost::property_tree::ptree*           tgModel
   )
{
   using boost::property_tree::ptree;
   
   try
   {
      ptree peLib;
      
      // Read in processing element library
      read_xml(tgModel->get<std::string>
	       ("system.constraints.pe_lib.<xmlattr>.file"), peLib);
      
      // Parse platform -> resource_list
      ptree node = tgModel->get_child("system.platform.resource_list");
      
      for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	  ++iter)
      {
	 // Confirm this is a resource tag
	 if((*iter).first == "resource")
	 {
	    // Check that its class is "pe" or not defined
	    std::string resourceClass = (*iter).second.get<std::string>
	       ("<xmlattr>.class", "pe");
	    // Read in ID
	    unsigned long int id = (*iter).second.get<unsigned long int>
	       ("<xmlattr>.id");
	    // Read name or give a default name of "<class> <id>"
	    std::string name = (*iter).second.get<std::string>
	       ("<xmlattr>.name", resourceClass + " " + 
		boost::lexical_cast<std::string>(id));	      
	    
	    if(resourceClass == "pe")
	    {
	       // Create new PE (which is a thread)
	       sctg::ProcessingElement* pe = new 
		  sctg::ProcessingElement(name.c_str(), 
					  (*iter).second, peLib, *cfg);
	       cfg->addResourceMap(pe->getId(), pe);
	       pes->push_back(pe);
	    }
	    else if(resourceClass == "memory")
	    {
	       // Create new memory (which is a thread)
	       sctg::MemoryModel* me = new
		  sctg::MemoryModel(name.c_str(), (*iter).second, *cfg);
	       cfg->addResourceMap(me->getId(), me);
	       mems->push_back(me);
	    }
	    else
	    {
	       std::string err = "Unknown resource class " + resourceClass;
	       throw std::runtime_error(err.c_str());
	    }
	 }
      }





      // Parse Tasks and task connections
      node = tgModel->get_child("system.application.task_graph");
      for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	  ++iter)
      {
	 if((*iter).first == "task")
	 {
	    sctg::Task* task = new sctg::Task((*iter).second, *cfg);
	    cfg->addResourceUserMap(task->getId(), task);
	    tasks->push_back(task);
	 }
	 else if((*iter).first == "mem_area")
	 {
	    sctg::MemArea* memArea = new sctg::MemArea((*iter).second, *cfg);
	    cfg->addResourceUserMap(memArea->getId(), memArea);
	    memAreas->push_back(memArea);
	 }
	 else if((*iter).first == "port_connection")
	 {
	    unsigned long int src = (*iter).second.get<unsigned long int>
	       ("<xmlattr>.src");
	    unsigned long int dst = (*iter).second.get<unsigned long int>
	       ("<xmlattr>.dst");
	    cfg->addPortConnection(src, dst);
	 }
      }

      node = tgModel->get_child("system.application");
      for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	  ++iter)
      {
	 if((*iter).first == "port_connection")
	 {
	    unsigned long int src = (*iter).second.get<unsigned long int>
	       ("<xmlattr>.src");
	    unsigned long int dst = (*iter).second.get<unsigned long int>
	       ("<xmlattr>.dst");
	    cfg->addPortConnection(src, dst);
	 }
      }

      std::cout << std::endl;


      // Parse events
      node = tgModel->get_child("system.application.task_graph.event_list");
      for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	  ++iter)
      {
	 if((*iter).first == "event")
	 {
	    // Read in ID
	    unsigned long int id = (*iter).second.get<unsigned long int>
	       ("<xmlattr>.id");
	    // Read name or give a default name of "Event <id>"
	    std::string name = (*iter).second.get<std::string>
	       ("<xmlattr>.name", "Event " + 
		boost::lexical_cast<std::string>(id));	      
	    // Create new event (which is a thread)
	    sctg::Event* event = new sctg::Event(name.c_str(),
						 *cfg,
						 (*iter).second);	    
	    events->push_back(event);
	 }
      }


      // Map tasks to PEs and mem_areas to memories
      node = tgModel->get_child("system.mapping");
      for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	  ++iter)
      {
	 if((*iter).first == "resource")
	 {
	    unsigned long int resourceId = 
	       (*iter).second.get<unsigned long int>("<xmlattr>.id");
	    // Check for sw_platform
	    ptree groupNode = 
	       (*iter).second.get_child("sw_platform", (*iter).second);
	    // If sw_platform is not found use resource instead

	    // Loop over all groups
	    for(ptree::const_iterator iter2 = groupNode.begin();
		iter2 != groupNode.end(); ++iter2)
	    {
	       if((*iter2).first == "group")
	       {
		  unsigned long int groupId = 
		     (*iter2).second.get<unsigned long int>("<xmlattr>.id");
		  for(ptree::const_iterator iter3 = (*iter2).second.begin();
		      iter3 != (*iter2).second.end(); ++iter3)
		  {		     
		     if((*iter3).first == "task")
		     {
			unsigned long int taskId = 
			   (*iter3).second.get<unsigned long int>
			   ("<xmlattr>.id");			
			cfg->getResource
			   (resourceId)->mapResourceUser
			   (cfg->getResourceUser(taskId));
			cfg->addResourceUserToResourceMap
			   (taskId, cfg->getResource(resourceId));
			cfg->addTaskToGroup(taskId, groupId);
			cfg->addGroupToPe(groupId, resourceId);
		     }
		  }
	       }
	       else if((*iter2).first == "mem_area")
	       {
		  unsigned long int id = 
		     (*iter2).second.get<unsigned long int>("<xmlattr>.id");

		  cfg->getResource(resourceId)->mapResourceUser
		     (cfg->getResourceUser(id));
		  cfg->addResourceUserToResourceMap
		     (id, cfg->getResource(resourceId));
	       }
	    }

	 }
      }      

      // Map events to resources
      for(unsigned int i = 0; i < events->size(); ++i)
      {
	 for(unsigned int j = 0; j < pes->size(); ++j)
	 {	      
	    if(pes->at(j)->hasInPort
	       (cfg->getDestination(events->at(i)->getOutPort())))
	    {	
	       events->at(i)->mapPe(pes->at(j));
	    }
	 }
      }


      std::cout << std::endl;
      
      // Connect network interfaces to processing elements
      for(unsigned int i = 0; i < pes->size(); ++i)
      {
	 unsigned long int terminalId = pes->at(i)->getTerminal();
	 unsigned long int portId = 0;
	 unsigned long int routerId = 0;
	 unsigned long int width = 0;
	 unsigned long int address = 0;
	 unsigned long int frequency = 0;
	 // Search router and its port id
	 node = tgModel->get_child("system.platform.noc.terminal_list");	 
	 for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	     ++iter)
	 {
	    if((*iter).first == "connection")
	    {
	       if((*iter).second.get<unsigned long int>("<xmlattr>.id") 
		  == terminalId)
	       {
		  portId = (*iter).second.get<unsigned long int>
		     ("<xmlattr>.port");
		  routerId = (*iter).second.get<unsigned long int>
		     ("<xmlattr>.router");
	       }
	    }	    
	 }
	 node = tgModel->get_child("system.platform.noc.router_list");
	 for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	     ++iter)
	 {
	    // Search for correct router
	    if((*iter).first == "router" && 
	       (*iter).second.get<unsigned long int>("<xmlattr>.id") ==
	       routerId)
	    {
	       width = (*iter).second.get<unsigned long int>
		  ("<xmlattr>.width");
	       // Search correct port
	       for(ptree::const_iterator iter2 = (*iter).second.begin();
		   iter2 != (*iter).second.end(); ++iter2)
	       {		  
		  if((*iter2).first == "port" &&
		     (*iter2).second.get<unsigned long int>("<xmlattr>.id") ==
		     portId)
		  {		     
		     std::string addrStr = 
			(*iter2).second.get<std::string>
			("<xmlattr>.address");		     
		     std::istringstream iss(addrStr);
		     iss >> std::hex >> address;
		     break;
		  }
	       }
	       break;
	    }
	 }
	 pes->at(i)->setTerminal(terminalId, address, width);

	 std::cout << "Found address 0x" << std::hex << address
		   << std::dec << " for PE " << pes->at(i)->getName()
		   << " (terminal:" << terminalId << " router:"
		   << routerId << " port:" << portId << ")" << std::endl;
      }


      // Connect network interfaces to processing elements
      for(unsigned int i = 0; i < mems->size(); ++i)
      {
	 unsigned long int terminalId = mems->at(i)->getTerminal();
	 unsigned long int portId = 0;
	 unsigned long int routerId = 0;
	 unsigned long int width = 0;
	 unsigned long int address = 0;
	 unsigned long int frequency = 0;
	 // Search router and its port id
	 node = tgModel->get_child("system.platform.noc.terminal_list");	 
	 for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	     ++iter)
	 {
	    if((*iter).first == "connection")
	    {
	       if((*iter).second.get<unsigned long int>("<xmlattr>.id") 
		  == terminalId)
	       {
		  portId = (*iter).second.get<unsigned long int>
		     ("<xmlattr>.port");
		  routerId = (*iter).second.get<unsigned long int>
		     ("<xmlattr>.router");
	       }
	    }	    
	 }
	 node = tgModel->get_child("system.platform.noc.router_list");
	 for(ptree::const_iterator iter = node.begin(); iter != node.end(); 
	     ++iter)
	 {
	    // Search for correct router
	    if((*iter).first == "router" && 
	       (*iter).second.get<unsigned long int>("<xmlattr>.id") ==
	       routerId)
	    {
	       width = (*iter).second.get<unsigned long int>
		  ("<xmlattr>.width");
	       // Search correct port
	       for(ptree::const_iterator iter2 = (*iter).second.begin();
		   iter2 != (*iter).second.end(); ++iter2)
	       {		  
		  if((*iter2).first == "port" &&
		     (*iter2).second.get<unsigned long int>("<xmlattr>.id") ==
		     portId)
		  {		     
		     std::string addrStr = 
			(*iter2).second.get<std::string>
			("<xmlattr>.address");		     
		     std::istringstream iss(addrStr);
		     iss >> std::hex >> address;
		     break;
		  }
	       }
	       break;
	    }
	 }
	 mems->at(i)->setTerminal(terminalId, address, width);
	 
	 std::cout << "Found address 0x" << std::hex << address
		   << std::dec << " for PE " << mems->at(i)->getName()
		   << " (terminal:" << terminalId << " router:"
		   << routerId << " port:" << portId << ")" << std::endl;
      }
      
      std::cout << std::endl;


      std::string nocClass;
      std::string nocType;
      std::string nocSubType;
      node = tgModel->get_child("system.platform.noc");	 
      nocClass   = node.get<std::string>("<xmlattr>.class", "UNKNOWN");
      nocType    = node.get<std::string>("<xmlattr>.type", "UNKNOWN");
      nocSubType = node.get<std::string>("<xmlattr>.subtype", "UNKNOWN");

      cfg->setNocClass(nocClass);
      cfg->setNocType(nocType);
      cfg->setNocSubType(nocSubType);       



      // Basic validation for xml model. Methods on conf will throw exception 
      // somethign goes wrong
      
      for(std::vector<sctg::Task*>::iterator iter = tasks->begin();
	  iter != tasks->end(); ++iter)
      {
	 std::vector<unsigned long int> outs = (*iter)->getOutPorts();
	 for(int i = 0; i < outs.size(); ++i)
	 {
	    unsigned long int oport = outs.at(i);
	    unsigned long int iport = cfg->getDestination(oport);
	    sctg::Task* target = dynamic_cast<sctg::Task*>
	       (cfg->getResourceUserByInPort(iport));
	 }
	 std::vector<unsigned long int> ins = (*iter)->getInPorts();
	 for(int i = 0; i < ins.size(); ++i)
	 {
	    unsigned long int iport = ins.at(i);
	    unsigned long int oport = cfg->getSource(iport);
	    //sctg::Task* target = cfg->getTaskByInPort(iport);
	 }
      }



   }
   catch(std::exception& e)
   {
      std::cerr << "EXCEPTION: "
		<< e.what() << std::endl;
   }
   catch(...)
   {
      std::cerr << "EXCEPTION: unknown source (not std::exception)."
		<< std::endl;
   }



} // analyzeModel(...)









/**
 * Gather costs, such as PE utilization, execution counts, and transfer latencies. 
 * They are stored into CostFunction object.
 * 2011-08-19 This a new function but all the code is just moved from main to here.
 */
void calculateCost (sctg::Configuration* cfg, sctg::CostFunction* costF,
		    std::vector<sctg::ProcessingElement*>* pes,
		    std::vector<sctg::Task*>*              tasks,
		    std::vector<sctg::Event*>*             events
//   std::vector<sctg::MemArea*>           memAreas,
//   std::vector<sctg::MemoryModel*>       mems,

   )
{

   try 
   {
      std::ostringstream oss;
      std::string str;

      // PEs: utilization and frequency
      double avgPeUtil = 0.0;
      for(unsigned int i = 0; i < pes->size(); ++i)
      {
	 oss.str("");
	 oss << "pu_" << pes->at(i)->getId();
	 double util = pes->at(i)->getAvgUtilization();
	 avgPeUtil += util;
	 str = oss.str();
	 costF->addVariable(str, util);
	 oss.str("");
	 oss << "pu_" << pes->at(i)->getName();
	 str = oss.str();
	 costF->addVariable(str, util);

	 oss.str("");
	 oss << "pf_" << pes->at(i)->getId();
	 str = oss.str();
	 costF->addVariable(str, pes->at(i)->getFrequency());
      }
      str = "pu_avg";
      avgPeUtil /= pes->size();
      costF->addVariable(str, avgPeUtil);

      // Tasks and events: execution count
      double totalCount = 0.0;
      for(unsigned int i = 0; i < tasks->size(); ++i)
      {
	 oss.str("");
	 oss << "tc_" << tasks->at(i)->getId();
	 double count = tasks->at(i)->getTimesTriggered();
	 totalCount += count;
	 str = oss.str();
	 costF->addVariable(str, count);
	 oss.str("");
	 oss << "tc_" << tasks->at(i)->getName();
	 str = oss.str();
	 costF->addVariable(str, count);
      }
      str = "tc_tot";
      costF->addVariable(str, totalCount);

      totalCount = 0.0;
      for(unsigned int i = 0; i < events->size(); ++i)
      {
	 oss.str("");
	 oss << "ec_" << events->at(i)->getId();
	 double count = events->at(i)->getTimesHappened();
	 totalCount += count;
	 str = oss.str();
	 costF->addVariable(str, count);
	 oss.str("");
	 oss << "ec_" << events->at(i)->getName();
	 str = oss.str();
	 costF->addVariable(str, count);
      }
      str = "ec_tot";
      costF->addVariable(str, totalCount);
      
      // Tokens: latencies
      std::map<unsigned long int, 
	    std::map<unsigned long int, sc_core::sc_time> >::iterator outer;
      std::map<unsigned long int, sc_core::sc_time>::iterator inner;
      
      unsigned long int latcount = 0;
      double latavg = 0.0;
      
      for(outer = cfg->getTokenLatency().begin(); 
	  outer != cfg->getTokenLatency().end(); ++outer)
      {
	 for(inner = (*outer).second.begin(); 
	     inner != (*outer).second.end(); ++inner)
	 {
	    oss.str("");
	    oss << "latf_" << (*outer).first << "_" << (*inner).first << "_avg";
	    double val = 
	       ((*inner).second / sc_core::sc_time(1000.0, sc_core::SC_MS)) 
	       / cfg->getTokenCount()[(*outer).first][(*inner).first];
	    latcount++;
	    latavg += val;
	    str = oss.str();
	    costF->addVariable(str, val);
	    oss.str("");
	    oss << "latf_" << (*outer).first << "_" << (*inner).first << "_max";
	    val = cfg->getTokenLatencyMax()[(*outer).first][(*inner).first] / 
	       sc_core::sc_time(1000.0, sc_core::SC_MS);
	    str = oss.str();
	    costF->addVariable(str, val);
	    oss.str("");
	    oss << "latf_" << (*outer).first << "_" << (*inner).first << "_min";
	    val = cfg->getTokenLatencyMin()[(*outer).first][(*inner).first] / 
	       sc_core::sc_time(1000.0, sc_core::SC_MS);
	    str = oss.str();
	    costF->addVariable(str, val);
	 }
      }
      
      oss.str("");
      oss << "latf_tot_avg";
      str = oss.str();
      latavg /= latcount;
      costF->addVariable(str, latavg);
      
      // Calculate latencies with unsent tokens      
      latcount = 0;
      latavg = 0.0;
      
      for(unsigned int i = 0; i < pes->size(); ++i)
      {
	 pes->at(i)->updateUnfinishedTokensLatency();
      }
      
      for(outer = cfg->getTokenLatency().begin(); 
	  outer != cfg->getTokenLatency().end(); ++outer)
      {
	 for(inner = (*outer).second.begin(); 
	     inner != (*outer).second.end(); ++inner)
	 {
	    oss.str("");
	    oss << "lat_" << (*outer).first << "_" << (*inner).first << "_avg";
	    double val = 
	       ((*inner).second / sc_core::sc_time(1000.0, sc_core::SC_MS)) 
	       / cfg->getTokenCount()[(*outer).first][(*inner).first];
	    latcount++;
	    latavg += val;
	    str = oss.str();
	    costF->addVariable(str, val);
	    oss.str("");
	    oss << "lat_" << (*outer).first << "_" << (*inner).first << "_max";
	    val = cfg->getTokenLatencyMax()[(*outer).first][(*inner).first] / 
	       sc_core::sc_time(1000.0, sc_core::SC_MS);
	    str = oss.str();
	    costF->addVariable(str, val);
	    oss.str("");
	    oss << "lat_" << (*outer).first << "_" << (*inner).first << "_min";
	    val = cfg->getTokenLatencyMin()[(*outer).first][(*inner).first] / 
	       sc_core::sc_time(1000.0, sc_core::SC_MS);
	    str = oss.str();
	    costF->addVariable(str, val);
	 }
      }
      
      oss.str("");
      oss << "lat_tot_avg";
      str = oss.str();
      latavg /= latcount;
      costF->addVariable(str, latavg);
      
      // Path measurements: latencies
      for(std::map<std::string, sc_core::sc_time>::iterator iter =
	     cfg->getTotPathLat().begin(); iter != cfg->getTotPathLat().end();
	  ++ iter )
      {
	 oss.str("");
	 oss << "path_" << (*iter).first << "_avg";
	 str = oss.str();
	 double val = ((*iter).second / sc_core::sc_time(1000.0, sc_core::SC_MS)) /
	    cfg->getPathCount()[(*iter).first];
	 costF->addVariable(str, val);
	 oss.str("");
	 oss << "path_" << (*iter).first << "_count";
	 str = oss.str();
	 val = cfg->getPathCount()[(*iter).first];
	 costF->addVariable(str, val);
      }
      for(std::map<std::string, sc_core::sc_time>::iterator iter =
	     cfg->getMaxPathLat().begin(); iter != cfg->getMaxPathLat().end();
	  ++ iter )
      {
	 oss.str("");
	 oss << "path_" << (*iter).first << "_max";
	 str = oss.str();
	 double val = ((*iter).second / sc_core::sc_time(1000.0, sc_core::SC_MS));
	 costF->addVariable(str, val);
      }
      for(std::map<std::string, sc_core::sc_time>::iterator iter =
	     cfg->getMinPathLat().begin(); iter != cfg->getMinPathLat().end();
	  ++ iter )
      {
	 oss.str("");
	 oss << "path_" << (*iter).first << "_min";
	 str = oss.str();
	 double val = ((*iter).second / sc_core::sc_time(1000.0, sc_core::SC_MS));
	 costF->addVariable(str, val);
      }
      
      // Add misc cost function variables
      for(std::map<std::string, double>::iterator iter = 
	     cfg->getCostVariables().begin(); 
	  iter != cfg->getCostVariables().end(); ++iter)
      {
	 std::string ss((*iter).first);
	 double dd = (*iter).second;
	 costF->addVariable(ss, dd);
      }

   }
   catch(std::exception& e)
   {
      std::cerr << "EXCEPTION: "
		<< e.what() << std::endl;
   }
   catch(...)
   {
      std::cerr << "EXCEPTION: unknown source (not std::exception)."
		<< std::endl;
   }

      
} // calculateCost(...)


/** 
 * Print cost values to stdout and file 
 * 2011-08-19 This a new function but all the code is just moved from main to here.
 */
void printCost (sctg::Configuration* cfg, sctg::CostFunction* costF)
{

   try
   {

      unsigned int n = 1;

      for(std::vector<std::string>::iterator iter = 
	     cfg->getCostFunctions().begin(); 
	  iter != cfg->getCostFunctions().end(); ++iter)
      {
	 double answer = costF->parse((*iter));
	 
	 // To screen (one value)
	 std::cout << std::setiosflags(std::ios::left)
		   << "Cost " << std::setw(3) 
		   << n << " ; " << std::setprecision(16) 
		   << std::setw(22) << answer
		   << std::resetiosflags(std::ios::left)
		   << "   ; \"" << (*iter)
		   << "\"" << std::endl;
	 if(n % 5 == 0) {std::cout << std::endl;}
	 
	 // To summary log file (two values: regular and scientific notation)
	 if(cfg->getSummaryStream())
	 {
	    **(cfg->getSummaryStream())
	       << std::setiosflags(std::ios::fixed)
	       << "Cost " << std::setw(3) 
	       << n << " ; " << std::setprecision(16) 
	       << std::setw(26) << answer
	       << std::resetiosflags(std::ios::fixed)
	       << std::setiosflags(std::ios::scientific)
	       << " ; " << std::setprecision(16) 
	       << std::setw(26) << answer
	       << "   ; \"" << (*iter)
	       << "\"" << std::endl;    
	 }
	 ++n;
      }
      
   }
   catch(std::exception& e)
   {
      std::cerr << "EXCEPTION: "
		<< e.what() << std::endl;
   }
   catch(...)
   {
      std::cerr << "EXCEPTION: unknown source (not std::exception)."
		<< std::endl;
   }
   
} // printCost(...)


// Local Variables:
// mode: c++
// c-file-style: "ellemtel"
// c-basic-offset: 3
// End:
